import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from statistics import mean

from nltk.corpus import stopwords
from sklearn.metrics import classification_report, confusion_matrix

from tensorflow.keras.preprocessing import sequence
from tensorflow.keras.preprocessing.text import Tokenizer
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import LSTM, Dense, Dropout, Bidirectional, SpatialDropout1D, Conv1D, GlobalMaxPooling1D, Embedding
from tensorflow.keras.utils import to_categorical

#download stop words vocab - added by MR
# import nltk
# nltk.download('stopwords')-


# Load data
test = pd.read_csv("./data/test.txt", sep=";", header=None)
train = pd.read_csv("./data/train.txt", sep=";", header=None)
val = pd.read_csv("./data/val.txt", sep=";", header=None)

train.head()

# Split data to X (text) and Y (label)
x_train, y_train = train[0], train[1]
x_test, y_test = test[0], test[1]
x_val, y_val = val[0], val[1]

x_test[:10]

# Removing stop words
stop_words = stopwords.words('english')

for stop_word in stop_words:
    regex_stopword = r"\b" + stop_word + r"\b"
    x_train = x_train.str.replace(regex_stopword, "")
    x_test = x_test.str.replace(regex_stopword, "")
    x_val = x_val.str.replace(regex_stopword, "")

x_test[:10]

# Tokenize words
tok = Tokenizer()
text = np.concatenate((x_train, x_test, x_val), axis=0)
tok.fit_on_texts(text)  # Tokenize all text

# Tokenize each set
x_train = tok.texts_to_sequences(x_train)
x_test = tok.texts_to_sequences(x_test)
x_val = tok.texts_to_sequences(x_val)

n_words = len(tok.word_index) + 1

text_len = [len(x) for x in text]
print(f"Max len = {max(text_len)}\nMin len = {min(text_len)}\nAvg len = {mean(text_len)}")

# Replace str labels with numeric
labels = {
    "joy": 0,
    "sadness": 1,
    "anger": 2,
    "fear": 3,
    "love": 4,
    "surprise": 5
}

y_train = y_train.replace(labels)
y_val = y_val.replace(labels)
y_test = y_test.replace(labels)

# Normalize sequences to same size
vec_len = 50
x_train = sequence.pad_sequences(x_train, maxlen=vec_len)
x_test = sequence.pad_sequences(x_test, maxlen=vec_len)
x_val = sequence.pad_sequences(x_val, maxlen=vec_len)


# Convert labels to categorical
n_classes = len(np.unique(y_train))
y_train_encoded = to_categorical(y_train, n_classes)
y_test_encoded = to_categorical(y_test, n_classes)
y_val_encoded = to_categorical(y_val, n_classes)

print(f"Train: {x_train.shape} {y_train.shape}\nTest: {x_test.shape} {y_test.shape}\nVal: {x_val.shape} {y_val.shape}")

embedding_length = 128

model = Sequential()
model.add(Embedding(n_words, embedding_length, input_length=x_train.shape[1]))
model.add(SpatialDropout1D(0.5))

model.add(Bidirectional(LSTM(128, return_sequences=True, dropout=0.5, recurrent_dropout=0.5)))
model.add(Conv1D(128, 10, padding="valid", activation="relu", strides=3))
model.add(Dropout(0.5))

model.add(Bidirectional(LSTM(128, return_sequences=True, dropout=0.5, recurrent_dropout=0.5)))
model.add(Conv1D(128, 10, padding="valid", activation="relu", strides=3))
model.add(Dropout(0.5))
model.add(GlobalMaxPooling1D())

model.add(Dense(n_classes, activation='softmax'))

# Train Model
model.compile(loss="categorical_crossentropy", optimizer="adam", metrics=['accuracy'])
model_trained = model.fit(x_train, y_train_encoded, epochs=30, batch_size=128, validation_data=(x_val, y_val_encoded), verbose=2)

# =============================================================================
# saving/loading model
# =============================================================================
model.save("my_model.h5") #using h5 extension
print("model saved!!!")

import tensorflow as tf 

model = tf.keras.models.load_model('my_model.h5')
model.summary()
# =============================================================================
# 
# =============================================================================


# Plot training progress
fig, axs = plt.subplots(1,2, figsize=(18,5))
axs[0].plot(model_trained.history['loss'])
axs[1].plot(model_trained.history['accuracy'])
axs[0].plot(model_trained.history['val_loss'])
axs[1].plot(model_trained.history['val_accuracy'])
axs[0].set_title("Loss over epoch", fontsize=20)
axs[0].set_ylabel("loss", fontsize=20)
axs[0].set_xlabel("epoch", fontsize=20)
axs[1].set_title("Accuracy over epoch", fontsize=20)
axs[1].set_ylabel("acc", fontsize=20)
axs[1].set_xlabel("epoch", fontsize=20)
axs[0].legend(["loss", "val_loss"], loc="center right", fontsize=15)
axs[1].legend(["accuracy", "val_accuracy"], loc="center right", fontsize=15)
plt.show()

# Predict classes
predictions = np.argmax(model.predict(x_test), axis=-1)
# Create confusion matrix and show metrics
cm = confusion_matrix(y_test, predictions)
# cm = pd.DataFrame(cm).to_string()
report = classification_report(y_test, predictions)

plt.figure(figsize = (10,7))
sns.heatmap(cm, annot=True)
plt.savefig("NN_conf_matrix.png")
plt.show()
plt.clf()
print(f"Accuracy:\n{report}")#\nConfusion Matrix:\n{cm}")

